import numpy

from snoop_tensor.base import Formatter

class TensorFormatter(Formatter):
    def __init__(self, property_name=False, properties=('shape', 'dtype', 'has_nan', 'has_inf')):
        super().__init__("ndarray", property_name, properties)

    def repr_shape(self, array):
        return str(array.shape)

    def repr_dtype(self, array):
        return str(array.dtype)

    def repr_has_nan(self, array):
        result = bool(numpy.isnan(array).any())
        if self.properties_name:
            return str(result)
        if result:
            return 'has_nan'
        return ''

    def repr_has_inf(self, tensor):
        result = bool(numpy.isinf(tensor).any())
        if self.properties_name:
            return str(result)
        if result:
            return 'has_inf'
        return ''



class Formatter:
    def __init__(self, disp_name, property_name=False, properties=[]):
        self.properties = properties
        self.properties_name = property_name
        self.disp_name = disp_name

    def __call__(self, tensor):
        prefix = self.disp_name + '<'
        suffix = '>'
        properties_str = ''
        for p in self.properties:
            new = ''
            if self.properties_name:
                new += p + '='
            if hasattr(self, 'repr_' + p):
                new += getattr(self, 'repr_' + p)(tensor)
            else:
                raise ValueError('Unknown tensor property')

            if properties_str != '' and len(new) > 0:
                properties_str += ', '
            properties_str += new

        return prefix + properties_str + suffix



import os
import time

class GPUMemoryWatcher:
    def __init__(self, gpus=None):
        from py3nvml import py3nvml
        py3nvml.nvmlInit()

        self.gpus = gpus or range(py3nvml.nvmlDeviceGetCount())

    def __call__(self, source, value):
        from py3nvml import py3nvml

        try:
            ret = []
            for gidx in self.gpus:
                handle = py3nvml.nvmlDeviceGetHandleByIndex(gidx)
                procs = py3nvml.nvmlDeviceGetComputeRunningProcesses(handle)
                try:
                    cproc = next(p for p in procs if p.pid == os.getpid())
                except StopIteration:
                    continue
                usage = self._print_size(cproc.usedGpuMemory)
                ret.append("{}: {}".format(gidx, usage))

            if len(ret) == 0:
                return 'gpu_mem', 'unused'
            else:
                return 'gpu_mem', ', '.join(ret)
        except:
            print("ERROR")
            import sys
            e = sys.exc_info()[0]
            return 'gpu_mem', str(e)

    def _print_size(self, size):
        suffix = ['B', 'KB', 'MB', 'GB']
        level = 0
        size = float(size)
        while size > 1024:
            size /= 1024
            level += 1
        return "%.2f%s" % (size, suffix[level])

    @staticmethod
    def create_watcher(*args):
        watcher = GPUMemoryWatcher(*args)
        return lambda source, value: watcher(source, value)

class TimeCostWatcher:
    def __init__(self):
        self.last_time = time.time()

    def __call__(self, source, value):
        new_time = time.time()
        duration = new_time - self.last_time
        self.last_time = new_time

        return 'time elapsed', "%.4fs" % duration

    @staticmethod
    def create_watcher(*args):
        watcher = TimeCostWatcher(*args)
        return lambda source, value: watcher(source, value)



from setuptools import setup, find_packages

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name='snoop-tensor',
    version="0.2",
    author='Jacob Zhong',
    author_email='cmpute@gmail.com',
    description="Tensor extension for snoop",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/cmpute/snoop-tensor',
    packages=find_packages(),
    install_requires=[
        'snoop',
    ],
    tests_require=[
        'pytest',
        'torch',
        'python-toolbox',
        'coverage',
        'snoop',
    ],
)



import torch

from snoop_tensor.base import Formatter


class TensorFormatter(Formatter):
    FLOATING_POINTS = ['float', 'double', 'half', 'complex128', 'complex32', 'complex64']

    def __init__(self, property_name=False, properties=('shape', 'dtype', 'device', 'requires_grad', 'has_nan', 'has_inf')):
        super().__init__("tensor", property_name, properties)

    def repr_shape(self, tensor):
        ret = ''
        if not hasattr(tensor, 'names') or not tensor.has_names():
            ret += str(tuple(tensor.shape))
        else:
            ret += '('
            for n, v in zip(tensor.names, tensor.shape):
                if n is not None:
                    ret += '{}={}, '.format(n, v)
                else:
                    ret += '{}, '.format(v)
            ret = ret[:-2] + ')'
        return ret

    def repr_dtype(self, tensor):
        dtype_str = str(tensor.dtype)
        dtype_str = dtype_str[len('torch.'):]
        return dtype_str

    def repr_device(self, tensor):
        return str(tensor.device)

    def repr_requires_grad(self, tensor):
        if self.properties_name:
            return str(tensor.requires_grad)
        if tensor.requires_grad:
            return 'grad'
        return ''

    def repr_has_nan(self, tensor):
        result = tensor.dtype in TensorFormatter.FLOATING_POINTS and bool(torch.isnan(tensor).any())
        if self.properties_name:
            return str(result)
        if result:
            return 'has_nan'
        return ''

    def repr_has_inf(self, tensor):
        result = tensor.dtype in TensorFormatter.FLOATING_POINTS and bool(torch.isinf(tensor).any())
        if self.properties_name:
            return str(result)
        if result:
            return 'has_inf'
        return ''

class LayerFormatter:
    def __init__(self):
        pass

    def __call__(self, layer):
        return type(layer).__name__



import snoop
import cheap_repr
import snoop.configuration

def register(verbose=False,
    torch_formatter=None, numpy_formatter=None,
    report_gpu_mem=False, report_time_cost=False
):
    try:
        import numpy
        from snoop_tensor.numpy import TensorFormatter
        numpy_formatter = numpy_formatter or TensorFormatter()
        cheap_repr.register_repr(numpy.ndarray)(lambda x, _: numpy_formatter(x))
    except ImportError:
        pass

    try:
        import torch
        from snoop_tensor.torch import TensorFormatter, LayerFormatter
        torch_formatter = torch_formatter or TensorFormatter()
        torch_layer_formatter = LayerFormatter()
        cheap_repr.register_repr(torch.Tensor)(lambda x, _: torch_formatter(x))
        cheap_repr.register_repr(torch.nn.Module)(lambda x, _: torch_layer_formatter(x))
    except ImportError:
        pass

    if report_gpu_mem:
        from snoop_tensor.profiler import GPUMemoryWatcher
        snoop.config.watch_extras += (GPUMemoryWatcher.create_watcher(report_gpu_mem),)
    if report_time_cost:
        from snoop_tensor.profiler import TimeCostWatcher
        snoop.config.watch_extras += (TimeCostWatcher.create_watcher(),)

    unwanted = {
        snoop.configuration.len_shape_watch,
        snoop.configuration.dtype_watch,
    }
    snoop.config.watch_extras = tuple(x for x in snoop.config.watch_extras if x not in unwanted)
